// This file was generated by ecm_setup_version(): DO NOT EDIT!

#ifndef KDEVELOP_PG_QT_VERSION_H
#define KDEVELOP_PG_QT_VERSION_H

#define KDEVELOP_PG_QT_VERSION_STRING "2.3.0"
#define KDEVELOP_PG_QT_VERSION_MAJOR 2
#define KDEVELOP_PG_QT_VERSION_MINOR 3
#define KDEVELOP_PG_QT_VERSION_PATCH 0
#define KDEVELOP_PG_QT_VERSION ((2<<16)|(3<<8)|(0))

#endif
